#!/usr/bin/perl -w
#
# MD5 implemented in AMD CAL IL.
#
# Author: Marc Bevand <bevand_m (at) epita.fr>
#

use strict;

# pick a strategy for implementing shifting
my $shift_strategy = "c";
my %map_shft_literals = ();
my $code = "";

# Returns word X[k] processed during the specified step (r7-r22, or r28-r29).
sub step_to_x_k
{
    my ($step, $interleave) = @_;
    my @elm = qw/xxxx yyyy zzzz wwww/;
    my $k;
    die "*bug*: step cannot be $step\n" if $step >= 64;
    if ($step < 16)      { $k = $step; }
    elsif ($step < 16*2) { $k = (5 * $step + 1) % 16; }
    elsif ($step < 16*3) { $k = (3 * $step + 5) % 16; }
    else                 { $k = (7 * $step) % 16; }
    if ($interleave and ($k <= 1)) {
	# for the interleaved MD5, X[0] and X[1] are actually r28 and r29
	return sprintf "r%i", 28 + $k;
    }
    return sprintf "r%i", 7 + $k;
}

# Returns the sine value T_i used in the specified step (cb1[?].?).
sub step_to_t_i
{
    my ($step) = @_;
    my @elm = qw/xxxx yyyy zzzz wwww/;
    die "*bug*: step cannot be $step\n" if $step >= 64;
    return sprintf "cb1[%i].%s", $step / 4, $elm[$step % 4];
}

# Returns the register corresponding to A,B,C,D (r0-r3 or r24-r27).
sub abcd_to_r0123
{
    my $interleave = $_[4];
    my %h1 = (A => "r0", B => "r1", C => "r2", D => "r3");
    my %h2 = (A => "r24", B => "r25", C => "r26", D => "r27");
    for (my $i = 0; $i < 4; $i++) {
	my $h = (!$interleave ? \%h1 : \%h2);
	die "*bug*: invalid variable: $_[$i]\n" if !exists($h->{$_[$i]});
	$_[$i] = $h->{$_[$i]};
    }
}

# Declares constant used by shifting operations.
sub declare_shft_literals
{
    my @literals = ();
    my @elm = qw/xxxx yyyy zzzz wwww/;
    for my $s (
	7, 12, 17, 22, # round 1
	5, 9, 14, 20,  # round 2
	4, 11, 16, 23, # round 3
	6, 10, 15, 21, # round 4
    ) {
	if ($shift_strategy eq 'a') {
	    push @literals, 32 - $s, $s;
	} elsif ($shift_strategy eq 'b') {
	    push @literals, 32 - $s, 2 ** $s;
	} elsif ($shift_strategy eq 'c') {
	    push @literals, 32 - $s, $s;
	} else {
	    die "*bug*: invalid shift strategy: $shift_strategy\n";
	}
    }
    $code .= sprintf "  ; shift literals\n";
    for (my $i = 0; $i < scalar(@literals); $i++) {
	my $v = $literals[$i];
	my $lx = sprintf "l%d", $i / 4;
	my $e = $elm[$i % 4];
	$map_shft_literals{$v} = "$lx.$e";
	if (!($i % 4)) {
	    $code .= sprintf "  dcl_literal %s, %d", $lx, $v;
	} elsif (($i % 4) != 3) {
	    $code .= sprintf ", %d", $v;
	} else {
	    $code .= sprintf ", %d\n", $v;
	}
    }
}

# Returns the literal having the specified value.
sub get_shft_literal
{
    my ($v) = @_;
    exists($map_shft_literals{$v}) or die "cannot find literal $v\n";
    return $map_shft_literals{$v};
}

# Rotates r5 with a given shift strategy, stores the output in r5. Multiple
# strategies are available:
# a) uses 3 instructions (RIGHT SHIFT by "32-s", LEFT SHIFT by "s",
#   OR the 2 results)
# b) uses 2 instructions (RIGHT SHIFT by "32-s", MAD with "2**s" and
#   "right shift result")
# c) uses 1 instruction (BITALIGN with "32-s")
sub execute_shift_strategy
{
    my ($s_lit, $s_lit_compl, $s_lit_power) = @_;
    if ($shift_strategy eq 'a') {
	$code .= <<EOF;
    ishl r6, r5, $s_lit
    ushr r5, r5, $s_lit_compl
    ior  r5, r6, r5
EOF
    } elsif ($shift_strategy eq 'b') {
	$code .= <<EOF;
    ushr r6, r5, $s_lit_compl
    umad r5, r5, $s_lit_power, r6
EOF
    } elsif ($shift_strategy eq 'c') {
	$code .= <<EOF;
    bitalign r5, r5, r5, $s_lit_compl
EOF
    } else {
	die "unknown shift strategy: $shift_strategy\n";
    }
}

# Implements a round 1 step: t = x + ((t + F(x,y,z) + X[k] + T_i) <<< s)
sub round1_step
{
    my ($step, $t, $x, $y, $z, $s, $interleave) = @_;
    my $x_k = step_to_x_k($step, $interleave);
    my $t_i = step_to_t_i($step);
    abcd_to_r0123($t, $x, $y, $z, $interleave);
    my $s_lit = get_shft_literal($s) if ($shift_strategy eq 'a');
    my $s_lit_compl = get_shft_literal(32 - $s);
    my $s_lit_power = get_shft_literal(2 ** $s) if ($shift_strategy eq 'b');
    $code .= "\n".
    "    ; step $step (round 1)\n".
    # F(x,y,z)
    "    ixor r5, $y, $z\n".
    "    and  r5, r5, $x\n".
    "    ixor r5, r5, $z\n".
    # X[k] + T_i
    "    iadd r6, $x_k, $t_i\n".
    # sum = t + F() + X[k] + T_i
    "    iadd r5, $t, r5\n".
    "    iadd r5, r5, r6\n";
    # sum <<< s
    execute_shift_strategy($s_lit, $s_lit_compl, $s_lit_power);
    # add x, store in t
    $code .= "    iadd $t, $x, r5\n";
}

# Implements a round 2 step: t = x + ((t + G(x,y,z) + X[k] + T_i) <<< s)
sub round2_step
{
    my ($step, $t, $x, $y, $z, $s, $interleave) = @_;
    my $x_k = step_to_x_k($step, $interleave);
    my $t_i = step_to_t_i($step);
    abcd_to_r0123($t, $x, $y, $z, $interleave);
    my $s_lit = get_shft_literal($s) if ($shift_strategy eq 'a');
    my $s_lit_compl = get_shft_literal(32 - $s);
    my $s_lit_power = get_shft_literal(2 ** $s) if ($shift_strategy eq 'b');
    $code .= "\n".
    "    ; step $step (round 2)\n".
    # G(x,y,z)
    "    ixor r5, $x, $y\n".
    "    and  r5, r5, $z\n".
    "    ixor r5, r5, $y\n".
    # X[k] + T_i
    "    iadd r6, $x_k, $t_i\n".
    # sum = t + G() + X[k] + T_i
    "    iadd r5, $t, r5\n".
    "    iadd r5, r5, r6\n";
    # sum <<< s
    execute_shift_strategy($s_lit, $s_lit_compl, $s_lit_power);
    # add x, store in t
    $code .= "    iadd $t, $x, r5\n";
}

# Implements a round 3 step: t = x + ((t + H(x,y,z) + X[k] + T_i) <<< s)
sub round3_step
{
    my ($step, $t, $x, $y, $z, $s, $interleave) = @_;
    my $x_k = step_to_x_k($step, $interleave);
    my $t_i = step_to_t_i($step);
    abcd_to_r0123($t, $x, $y, $z, $interleave);
    my $s_lit = get_shft_literal($s) if ($shift_strategy eq 'a');
    my $s_lit_compl = get_shft_literal(32 - $s);
    my $s_lit_power = get_shft_literal(2 ** $s) if ($shift_strategy eq 'b');
    $code .= "\n".
    "    ; step $step (round 3)\n".
    # H(x,y,z)
    "    ixor r5, $y, $z\n".
    "    ixor r5, r5, $x\n".
    # X[k] + T_i
    "    iadd r6, $x_k, $t_i\n".
    # sum = t + H() + X[k] + T_i
    "    iadd r5, $t, r5\n".
    "    iadd r5, r5, r6\n";
    # sum <<< s
    execute_shift_strategy($s_lit, $s_lit_compl, $s_lit_power);
    # add x, store in t
    $code .= "    iadd $t, $x, r5\n";
}

# No need to implement a round 4 step:
# t = x + ((t + I(x,y,z) + X[k] + T_i) <<< s)
# Instead the reverse of a round 4 step is needed for reversing:
# t = ((t - x) >>> s) - I(x,y,z) - X[k] - T_i
sub rev_round4_step
{
    my ($step, $t, $x, $y, $z, $s, $interleave) = @_;
    my $x_k = step_to_x_k($step, $interleave);
    my $t_i = step_to_t_i($step);
    abcd_to_r0123($t, $x, $y, $z, $interleave);
    my $s_lit = get_shft_literal($s) if ($shift_strategy eq 'a' or
	$shift_strategy eq 'c');
    my $s_lit_compl = get_shft_literal(32 - $s);
    my $s_lit_power = get_shft_literal(2 ** $s) if ($shift_strategy eq 'b');
    # TODO: call execute_shift_strategy() instead of hardcoding bitalign
    $code .= "\n".
    "    ; reverse step $step (round 4)\n".
    # t - x
    "    inegate r5, $x\n".
    "    iadd r5, $t, r5\n".
    # rotate to the right
    "    bitalign r5, r5, r5, $s_lit\n".
    # I(x,y,z)
    "    inot r6, $z\n".
    "    ior r6, $x, r6\n".
    "    ixor r6, $y, r6\n".
    # subsract the value of I(...)
    "    inegate r6, r6\n".
    "    iadd r5, r5, r6\n".
    # substract X[k]
    "    inegate r6, $x_k\n".
    "    iadd r5, r5, r6\n".
    # substract T_i, storing the final value in t
    "    inegate r6, $t_i\n".
    "    iadd $t, r5, r6\n";
}

my $output = shift;
open STDOUT,">$output" or die "can't open $output: $!";
$code .= <<EOF;
il_cs
  dcl_num_thread_per_group %d
  ; message blocks
  dcl_cb cb0[4]
  ; sine values
  dcl_cb cb1[16]
EOF
declare_shft_literals();
$code .= <<EOF;
  ; l100.x number of iterations
  ; l100.y used to access g[], must be sizeof (thread_state_t) / 16
  ; l100.z -1, used in various places
  ; l100.w 1, used in various places
  dcl_literal l100, %d, %lu, -1, 1
  ; l101.xyzw initial intermediate hash values A, B, C, D
  dcl_literal l101, 0x67452301, 0xefcdab89, 0x98badcfe, 0x10325476
  ; l102.xyzw target hash A, B, C, D
  dcl_literal l102, 0x%x, 0x%x, 0x%x, 0x%x
  ; l103.x mask to extract position in pos_wuid
  ; l103.y 0, used in various places
  ; l103.z 2, used in various places
  ; l103.w value of "s_nomore"
  dcl_literal l103, 0x%x, 0, 2, %d

  ; r0-r3   A,B,C,D for first interleaved MD5
  ; r4.x    counter
  ; r4.y    offset to this thread's state in g[]
  ; r5-r6   temporary values
  ; r7-r8   message words X[0]-X[1] for first interleaved MD5
  ; r9-r22  message words X[2]-X[15]
  ; r23     target value D at step 45 for first interleaved MD5
  ; r24-r27 A,B,C,D for second interleaved MD5
  ; r28-r29 message words X[0]-X[1] for second interleaved MD5
  ; r30     target value D at step 45 for second interleaved MD5

  umul r4.y, vAbsTidFlat.x, l100.y

  ; load X[0]-X[1] for first interleaved MD5
  mov  r7.x, g[r4.y+0].y
  mov  r8.x, g[r4.y+0].z
  mov  r7.y, g[r4.y+1].x
  mov  r8.y, g[r4.y+1].y
  mov  r7.z, g[r4.y+1].w
  mov  r8.z, g[r4.y+2].x
  mov  r7.w, g[r4.y+2].z
  mov  r8.w, g[r4.y+2].w
  ; load X[0]-X[1] for second interleaved MD5
  mov  r28.x, g[r4.y+3].y
  mov  r29.x, g[r4.y+3].z
  mov  r28.y, g[r4.y+4].x
  mov  r29.y, g[r4.y+4].y
  mov  r28.z, g[r4.y+4].w
  mov  r29.z, g[r4.y+5].x
  mov  r28.w, g[r4.y+5].z
  mov  r29.w, g[r4.y+5].w
  ; init X[2]-X[15]
  mov  r9, cb0[0].zzzz
  mov r10, cb0[0].wwww
  mov r11, cb0[1].xxxx
  mov r12, cb0[1].yyyy
  mov r13, cb0[1].zzzz
  mov r14, cb0[1].wwww
  mov r15, cb0[2].xxxx
  mov r16, cb0[2].yyyy
  mov r17, cb0[2].zzzz
  mov r18, cb0[2].wwww
  mov r19, cb0[3].xxxx
  mov r20, cb0[3].yyyy
  mov r21, cb0[3].zzzz
  mov r22, cb0[3].wwww

  ; Reverse MD5 steps to compute D after step 45, store D in r23.
  ;
  ; Given a message block of 16 32-bit words (x) where all the words are known
  ; except x[0], and a 128-bit hash value (hash), executes the last 15 steps of
  ; round 4 of MD5 in reverse (from step 63 down to and including step 49) to
  ; compute the known intermediate hash values from the oldest step possible.
  ; Out of these values, D comes from step 45 and is output into d_res. This
  ; function enables an MD5 brute forcer to be able to determine after executing
  ; only 46 steps (0-45) if the D value computed so far indicates a potential 
  ; candidate message match.

  ; Load A,B,C,D for first interleaved MD5
  mov r0, l102.xxxx
  mov r1, l102.yyyy
  mov r2, l102.zzzz
  mov r3, l102.wwww
  inegate r5, l101.xxxx
  iadd r0, r0, r5
  inegate r5, l101.yyyy
  iadd r1, r1, r5
  inegate r5, l101.zzzz
  iadd r2, r2, r5
  inegate r5, l101.wwww
  iadd r3, r3, r5
  ; Load A,B,C,D for second interleaved MD5
  mov r24, r0
  mov r25, r1
  mov r26, r2
  mov r27, r3
EOF
my $s = 63;
rev_round4_step($s  ,'B','C','D','A','21'); # step 63
rev_round4_step($s--,'B','C','D','A','21', 1);
rev_round4_step($s,  'C','D','A','B','15');
rev_round4_step($s--,'C','D','A','B','15', 1);
rev_round4_step($s,  'D','A','B','C','10');
rev_round4_step($s--,'D','A','B','C','10', 1);
rev_round4_step($s,  'A','B','C','D', '6');
rev_round4_step($s--,'A','B','C','D', '6', 1);
rev_round4_step($s,  'B','C','D','A','21');
rev_round4_step($s--,'B','C','D','A','21', 1);
rev_round4_step($s,  'C','D','A','B','15');
rev_round4_step($s--,'C','D','A','B','15', 1);
rev_round4_step($s,  'D','A','B','C','10');
rev_round4_step($s--,'D','A','B','C','10', 1);
rev_round4_step($s,  'A','B','C','D', '6');
rev_round4_step($s--,'A','B','C','D', '6', 1);
rev_round4_step($s,  'B','C','D','A','21');
rev_round4_step($s--,'B','C','D','A','21', 1);
rev_round4_step($s,  'C','D','A','B','15');
rev_round4_step($s--,'C','D','A','B','15', 1);
rev_round4_step($s,  'D','A','B','C','10');
rev_round4_step($s--,'D','A','B','C','10', 1);
rev_round4_step($s,  'A','B','C','D', '6');
rev_round4_step($s--,'A','B','C','D', '6', 1);
rev_round4_step($s,  'B','C','D','A','21');
rev_round4_step($s--,'B','C','D','A','21', 1);
rev_round4_step($s,  'C','D','A','B','15');
rev_round4_step($s--,'C','D','A','B','15', 1);
rev_round4_step($s,  'D','A','B','C','10'); # step 49
rev_round4_step($s--,'D','A','B','C','10', 1);
# subsequent step (step 48) would need access to X[0] which is not known
$code .= <<EOF;
  ; save target values D for the two interleaved MD5
  mov r23, r3
  mov r30, r27

  ; init counter
  mov r4.x, l100.x

  whileloop
    ; break if counter is 0
    break_logicalz r4.x
    ; decrement loop counter
    iadd r4.x, r4.x, l100.z
    ; init A,B,C,D for two interleaved MD5
    mov r0, l101.xxxx
    mov r1, l101.yyyy
    mov r2, l101.zzzz
    mov r3, l101.wwww
    mov r24, r0
    mov r25, r1
    mov r26, r2
    mov r27, r3
EOF
$s = 0;

round1_step($s,  'A','B','C','D', '7');
round1_step($s++,'A','B','C','D', '7', 1);
round1_step($s,  'D','A','B','C','12');
round1_step($s++,'D','A','B','C','12', 1);
round1_step($s,  'C','D','A','B','17');
round1_step($s++,'C','D','A','B','17', 1);
round1_step($s,  'B','C','D','A','22');
round1_step($s++,'B','C','D','A','22', 1);
round1_step($s,  'A','B','C','D', '7');
round1_step($s++,'A','B','C','D', '7', 1);
round1_step($s,  'D','A','B','C','12');
round1_step($s++,'D','A','B','C','12', 1);
round1_step($s,  'C','D','A','B','17');
round1_step($s++,'C','D','A','B','17', 1);
round1_step($s,  'B','C','D','A','22');
round1_step($s++,'B','C','D','A','22', 1);
round1_step($s,  'A','B','C','D', '7');
round1_step($s++,'A','B','C','D', '7', 1);
round1_step($s,  'D','A','B','C','12');
round1_step($s++,'D','A','B','C','12', 1);
round1_step($s,  'C','D','A','B','17');
round1_step($s++,'C','D','A','B','17', 1);
round1_step($s,  'B','C','D','A','22');
round1_step($s++,'B','C','D','A','22', 1);
round1_step($s,  'A','B','C','D', '7');
round1_step($s++,'A','B','C','D', '7', 1);
round1_step($s,  'D','A','B','C','12');
round1_step($s++,'D','A','B','C','12', 1);
round1_step($s,  'C','D','A','B','17');
round1_step($s++,'C','D','A','B','17', 1);
round1_step($s,  'B','C','D','A','22');
round1_step($s++,'B','C','D','A','22', 1);

round2_step($s,  'A','B','C','D', '5');
round2_step($s++,'A','B','C','D', '5', 1);
round2_step($s,  'D','A','B','C', '9');
round2_step($s++,'D','A','B','C', '9', 1);
round2_step($s,  'C','D','A','B','14');
round2_step($s++,'C','D','A','B','14', 1);
round2_step($s,  'B','C','D','A','20');
round2_step($s++,'B','C','D','A','20', 1);
round2_step($s,  'A','B','C','D', '5');
round2_step($s++,'A','B','C','D', '5', 1);
round2_step($s,  'D','A','B','C', '9');
round2_step($s++,'D','A','B','C', '9', 1);
round2_step($s,  'C','D','A','B','14');
round2_step($s++,'C','D','A','B','14', 1);
round2_step($s,  'B','C','D','A','20');
round2_step($s++,'B','C','D','A','20', 1);
round2_step($s,  'A','B','C','D', '5');
round2_step($s++,'A','B','C','D', '5', 1);
round2_step($s,  'D','A','B','C', '9');
round2_step($s++,'D','A','B','C', '9', 1);
round2_step($s,  'C','D','A','B','14');
round2_step($s++,'C','D','A','B','14', 1);
round2_step($s,  'B','C','D','A','20');
round2_step($s++,'B','C','D','A','20', 1);
round2_step($s,  'A','B','C','D', '5');
round2_step($s++,'A','B','C','D', '5', 1);
round2_step($s,  'D','A','B','C', '9');
round2_step($s++,'D','A','B','C', '9', 1);
round2_step($s,  'C','D','A','B','14');
round2_step($s++,'C','D','A','B','14', 1);
round2_step($s,  'B','C','D','A','20');
round2_step($s++,'B','C','D','A','20', 1);

round3_step($s,  'A','B','C','D', '4');
round3_step($s++,'A','B','C','D', '4', 1);
round3_step($s,  'D','A','B','C','11');
round3_step($s++,'D','A','B','C','11', 1);
round3_step($s,  'C','D','A','B','16');
round3_step($s++,'C','D','A','B','16', 1);
round3_step($s,  'B','C','D','A','23');
round3_step($s++,'B','C','D','A','23', 1);
round3_step($s,  'A','B','C','D', '4');
round3_step($s++,'A','B','C','D', '4', 1);
round3_step($s,  'D','A','B','C','11');
round3_step($s++,'D','A','B','C','11', 1);
round3_step($s,  'C','D','A','B','16');
round3_step($s++,'C','D','A','B','16', 1);
round3_step($s,  'B','C','D','A','23');
round3_step($s++,'B','C','D','A','23', 1);
round3_step($s,  'A','B','C','D', '4');
round3_step($s++,'A','B','C','D', '4', 1);
round3_step($s,  'D','A','B','C','11');
round3_step($s++,'D','A','B','C','11', 1);
round3_step($s,  'C','D','A','B','16');
round3_step($s++,'C','D','A','B','16', 1);
round3_step($s,  'B','C','D','A','23');
round3_step($s++,'B','C','D','A','23', 1);
round3_step($s,  'A','B','C','D', '4');
round3_step($s++,'A','B','C','D', '4', 1);
round3_step($s,  'D','A','B','C','11');
round3_step($s++,'D','A','B','C','11', 1);
# can stop after step 45 (out of 0..63) since we have computed D.
# the next steps in round 3 would have been:
#round3_step($s,  'C','D','A','B','16');
#round3_step($s++,'C','D','A','B','16', 1);
#round3_step($s,  'B','C','D','A','23');
#round3_step($s++,'B','C','D','A','23', 1);

$code .= <<EOF;

    ; if first interleaved MD5 found the target value D, set bits in r5
    ieq r5, r3, r23

    ; if second interleaved MD5 found the target value D, set bits in r5
    ieq r6, r27, r30
    ior r5, r5, r6

    ; increment X[0] for first interleaved MD5, if one of the elms completed
    ; its work unit, set bits in r5
    iadd r7, r7, l100.wwww
    iand r6, r7, l103.xxxx
    ieq r6, r6, l103.yyyy
    ior r5, r5, r6

    ; increment X[0] for second interleaved MD5, if one of the elms completed
    ; its work unit, set bits in r5
    iadd r28, r28, l100.wwww
    iand r6, r28, l103.xxxx
    ieq r6, r6, l103.yyyy
    ior r5, r5, r6

    ; if r5 has any bit set, break
    ior r5.xy, r5.xy, r5.zw
    ior r5.x, r5.x, r5.y
    break_logicalnz r5.x
  endloop

EOF

my @components = qw/x y z w/;
# build the names of the 'sizeof (thread_state_t) / 4' words of state that
# are going to be written to g[]
my @gwords = ();
for (my $i = 0; $i < 24; $i++) {
    push @gwords, sprintf("g[r4.y+%d].%s", $i / 4, $components[$i % 4]);
}

$code .= <<EOF;
  ; analyze results of first interleaved MD5, and
  ; store the 4 elements' states in global buffer
  ieq r5, r3, r23        ; found D -> set bits in r5
  iand r6, r7, l103.xxxx
  ieq r6, r6, l103.yyyy  ; completed work unit -> set bits in r6

EOF
foreach my $elm (0 .. 3) {
    my $compo = $components[$elm];
    my $status = shift @gwords;
    my $wuid0 = shift @gwords;
    my $wuid1 = shift @gwords;
    $code .= <<EOF;
  ; elm $elm ($compo)
  ixor r0.x, $status, l103.w
  if_logicalnz r0.x
    mov $status, l103.y
    if_logicalnz r5.$compo
      mov $status, l100.w
    else
      if_logicalnz r6.$compo
        mov $status, l103.z
      endif
    endif
    mov $wuid0, r7.$compo
    mov $wuid1, r8.$compo
  endif

EOF
}

$code .= <<EOF;
  ; analyze results of second interleaved MD5, and
  ; store the 4 elements' states in global buffer
  ieq r5, r27, r30       ; found D -> set bits in r5
  iand r6, r28, l103.xxxx
  ieq r6, r6, l103.yyyy  ; completed work unit -> set bits in r6

EOF
foreach my $elm (0 .. 3) {
    my $compo = $components[$elm];
    my $status = shift @gwords;
    my $wuid0 = shift @gwords;
    my $wuid1 = shift @gwords;
    $code .= <<EOF;
  ; elm $elm ($compo)
  ixor r0.x, $status, l103.w
  if_logicalnz r0.x
    mov $status, l103.y
    if_logicalnz r5.$compo
      mov $status, l100.w
    else
      if_logicalnz r6.$compo
        mov $status, l103.z
      endif
    endif
    mov $wuid0, r28.$compo
    mov $wuid1, r29.$compo
  endif

EOF
}

$code .= "end\n";
# $code contains a compute shader program:
#   il_cs
#     ...
#   end
# now it is transformed to a C macro string:
#   #define KERNEL \
#   "il_cs\
#     ...\
#   end\
#   "
chomp $code;
$code =~ s/"/\\"/g;
$code =~ s/$/\\n\\/gm;
$code = "#define KERNEL \\\n\"$code\n\"\n";
print $code;

# eof
