#define _POSIX_C_SOURCE 199309L
#define _GNU_SOURCE
#include <assert.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <cal.h>
#include <calcl.h>
#include <stdbool.h>
#include <stdint.h>

#include "cal-utils.h"
#include "cpu-md5.h"
#include "kernel-md5.h"

unsigned max_gpus = 0;
unsigned pw_length = 5;
int threads_per_grp = 256;
int verbose = 0;
int disassemble_target = -1;
int iterations = 0x10000;
uint32_t target_hash[4] = { 0 };
uint8_t *message_block = NULL;
uint64_t first_wuid;
uint64_t last_wuid;
uint64_t next_wuid; // next work unit ID (without the position bits)
uint32_t sine_values[] = {
    0xd76aa478, 0xe8c7b756, 0x242070db, 0xc1bdceee,
    0xf57c0faf, 0x4787c62a, 0xa8304613, 0xfd469501,
    0x698098d8, 0x8b44f7af, 0xffff5bb1, 0x895cd7be,
    0x6b901122, 0xfd987193, 0xa679438e, 0x49b40821,
    0xf61e2562, 0xc040b340, 0x265e5a51, 0xe9b6c7aa,
    0xd62f105d,  0x2441453, 0xd8a1e681, 0xe7d3fbc8,
    0x21e1cde6, 0xc33707d6, 0xf4d50d87, 0x455a14ed,
    0xa9e3e905, 0xfcefa3f8, 0x676f02d9, 0x8d2a4c8a,
    0xfffa3942, 0x8771f681, 0x6d9d6122, 0xfde5380c,
    0xa4beea44, 0x4bdecfa9, 0xf6bb4b60, 0xbebfbc70,
    0x289b7ec6, 0xeaa127fa, 0xd4ef3085,  0x4881d05,
    0xd9d4d039, 0xe6db99e5, 0x1fa27cf8, 0xc4ac5665,
    0xf4292244, 0x432aff97, 0xab9423a7, 0xfc93a039,
    0x655b59c3, 0x8f0ccc92, 0xffeff47d, 0x85845dd1,
    0x6fa87e4f, 0xfe2ce6e0, 0xa3014314, 0x4e0811a1,
    0xf7537e82, 0xbd3af235, 0x2ad7d2bb, 0xeb86d391,
};

// values defined for "status" field below
const uint8_t s_searching = 0; // still searching in the work unit
const uint8_t s_found = 1;     // found a potential match
const uint8_t s_finished = 2;  // finished the work unit
const uint8_t s_nomore = 3;    // no more work units to give to this thread

// The pos_lsb less significant bits of the 64-bit "pos_wuid" field below
// represent the position in the work unit. The remaining 64-pos_lsb bits
// are the wuid (work unit id).
const int pos_lsb = 16;

typedef struct
{
    uint8_t	status;
    uint8_t	_unused[3];
    uint64_t	pos_wuid; // position in work unit, and work unit identifier
} __attribute__((packed))	elm_state_t;

typedef struct
{
    // each thread works on 8 IL elements (2 pairs of x,y,z,w)
#define ELM_PER_THREAD	8
    elm_state_t	elm[ELM_PER_THREAD];
} __attribute__((packed))	thread_state_t;

typedef struct
{
    unsigned		nr_simds;
    int			nr_threads;
    CALobject		obj;
    CALimage		img;
    CALdevice		device;
    CALcontext		ctx;
    CALmodule		module;
    CALfunc		entry;
    CALresource		globalRes;
    CALmem		globalMem;
    CALresource		msgRes;
    CALmem		msgMem;
    CALresource		sineRes;
    CALmem		sineMem;
    CALprogramGrid	pg;
    CALevent		e;
    bool		have_run;
    struct timeval	tv_start;
    struct timeval	tv_end;
    int			last_mhashpsec;
    int			nr_elms_done;
    bool		all_done;
}		gpu_state_t;

void init_md5_globals(void)
{
    message_block = malloc(64);
    memset(message_block, 0, 64);
    // append a 1 bit and save the 64-bit bit-length at the end of the block
    message_block[pw_length] = 0x80;
    uint64_t *bitlength = (uint64_t*)(message_block + 64 - 8);
    *bitlength = 8 * pw_length;
    // the 1 bit added to the message as a byte (0x80) plus one is 0x81.
    // if this work unit ID is ever assigned it means we have exhausted the
    // search space.
    first_wuid = 0x80ULL << (pw_length * 8 - pos_lsb);
    last_wuid = 0x81ULL << (pw_length * 8 - pos_lsb);
    next_wuid = first_wuid;
}

/*
 * Returns the next work unit, and set the status to "searching" if one
 * was assigned, or "no more" if units to process have been exhausted.
 */
uint64_t get_next_pos_wuid(uint8_t *status)
{
    if (next_wuid >= last_wuid)
	*status = s_nomore;
    else
	*status = s_searching;
    return (next_wuid++) << pos_lsb;
}

void print_hex(char *msg, unsigned length)
{
    printf("[");
    for (unsigned i = 0; i < length; i++)
      {
	if (i > 0)
	    printf(" ");
	printf("%02x", (uint8_t)msg[i]);
      }
    printf("]\n");
}

void print_str(char *msg, unsigned length)
{
    printf("[");
    for (unsigned i = 0; i < length; i++)
	putchar(msg[i]);
    printf("]\n");
}

/*
 * Prints candidate message to stdout.
 */
void found_candidate(CALuint devi, int t, int e, char *msg, unsigned length)
{
    printf("\n");
    if (verbose)
	printf("Message found by GPU %d thread %d elm %d\n", devi, t, e);
    printf("Found message.\nLength: %u\n", length);
    printf("Bytes:  ");
    print_hex(msg, length);
    printf("String: ");
    print_str(msg, length);
}

/*
 * Validates the candidate message, prints to stdout if it is a match.
 *
 * Returns true iff it is a true positive match, false if false positive.
 */
bool validate_candidate(CALuint devi, int t, int e, uint64_t pos_wuid)
{
    char candidate_message[64];
    MD5_CTX ctx;
    unsigned char candidate_hash[16];
    memcpy(candidate_message, message_block, sizeof (candidate_message));
    memcpy(candidate_message, &pos_wuid, sizeof (pos_wuid));
    MD5_Init(&ctx);
    MD5_Update(&ctx, candidate_message, pw_length);
    MD5_Final(candidate_hash, &ctx);
    if (!memcmp(target_hash, candidate_hash, sizeof (target_hash)))
      {
	found_candidate(devi, t, e, candidate_message, pw_length);
	return true;
      }
    return false;
}

void load_target_values(const char *h)
{
    if (32 != strlen(h))
	fprintf(stderr, "Target hash should be a hexadecimal representation of "
	       "a 128-bit MD5 hash:\n%s\n", h), exit(1);
    for (int i = 0; i < 16; i++)
      {
	char hex_byte[] = { h[i * 2], h[i * 2 + 1], 0 };
	((uint8_t *)target_hash)[i] = strtoul(hex_byte, NULL, 16);
      }
}

void show_global_stats(gpu_state_t *gs_base, CALuint nr_devs)
{
    CALuint devi;
    int global_mhashpsec = 0;
    for (devi = 0; devi < nr_devs; devi++)
	global_mhashpsec += gs_base[devi].last_mhashpsec;
    printf("Overall rate: %u Mhash/sec...", global_mhashpsec);
    if (verbose)
	printf("\n");
    else {
	printf("\r");
	fflush(stdout);
    }
}

void show_stats(CALuint devi, gpu_state_t *gs)
{
    long long ms0 = gs->tv_start.tv_sec * 1000 + gs->tv_start.tv_usec / 1000;
    long long ms1 = gs->tv_end.tv_sec * 1000 + gs->tv_end.tv_usec / 1000;
    int mhashpsec = (int)
	((float)ELM_PER_THREAD // nr of hashes verified per thread per iteration
	 * iterations // nr of iterations of the main loop for each thread
	 * gs->nr_threads // nr of threads
	 * 1000 / (ms1 - ms0) // hash/period_of_time converted to hash/sec
	 / 1e6); // converted to Mhash/sec
    if (verbose)
	printf("Device %d: execution time %lld ms (%u Mhash/sec)\n",
		devi, ms1 - ms0, mhashpsec);
    gs->last_mhashpsec = mhashpsec;
}

void set_local_res_mem(CALdevice device, CALcontext ctx, CALmodule module,
	CALresource *res, CALuint flags,
	CALmem *mem, const void *data, unsigned length, const char *param_name)
{
    /* About calResAllocLocal2D: There are some performance implications when
     * width is not a multiple of 64 for R6xx GPUs. Neither the width nor the
     * height can be too large (with 1.4 SDK on R700 the limit is 8192).
     * Similar limitations may exist for calResAllocLocal1D.
     */

    // allocate resource (length is in bytes)
    if (length % 64)
	fprintf(stderr, "length not multiple of 64: %u\n", length), exit(1);
    if (CAL_RESULT_OK != calResAllocLocal1D(res, device,
		length / 4, CAL_FORMAT_UNORM_INT32_1, flags))
	fatal("calResAllocLocal1D");
    if (CAL_RESULT_OK != calCtxGetMem(mem, ctx, *res))
	fatal("calCtxGetMem");
    // map, initialize, unmap
    if (data)
      {
	void *mapped;
	CALuint pitch;
	if (CAL_RESULT_OK != calResMap((CALvoid**)&mapped, &pitch, *res, 0))
	    fatal("calResMap");
	memcpy(mapped, data, length);
	if (CAL_RESULT_OK != calResUnmap(*res))
	    fatal("calResUnmap");
      }
    // bind to appropriate parameter
    CALname n;
    if (CAL_RESULT_OK != calModuleGetName(&n, ctx, module, param_name))
	fatal("calModuleGetName");
    if (CAL_RESULT_OK != calCtxSetMem(ctx, n, *mem))
	fatal("calCtxSetMem");
}

void prepare_run(CALuint devi, gpu_state_t *gs)
{
    // open device
    if (CAL_RESULT_OK != calDeviceOpen(&gs->device, devi))
	fatal("calDeviceOpen");
    if (CAL_RESULT_OK != calCtxCreate(&gs->ctx, gs->device))
	fatal("calCtxCreate");

    // load module, get entry point
    if (CAL_RESULT_OK != calModuleLoad(&gs->module, gs->ctx, gs->img))
	fatal("calModuleLoad");
    if (CAL_RESULT_OK != calModuleGetEntry(&gs->entry, gs->ctx, gs->module,
		"main"))
	fatal("calModuleGetEntry");

    // global buffer "g[]" (gs->nr_threads * 96 bytes)
    if (verbose)
	printf("Initializing global buffer\n");
    set_local_res_mem(gs->device, gs->ctx, gs->module,
	    &gs->globalRes, CAL_RESALLOC_GLOBAL_BUFFER,
	    &gs->globalMem, NULL,
	    gs->nr_threads * sizeof (thread_state_t), "g[]");

    // message block "cb0" (64 bytes)
    if (verbose)
	printf("Initializing message block\n");
    set_local_res_mem(gs->device, gs->ctx, gs->module,
	    &gs->msgRes, 0,
	    &gs->msgMem, message_block, 64, "cb0");

    // sine values "cb1" (64 4-byte values)
    if (verbose)
	printf("Initializing sine values\n");
    set_local_res_mem(gs->device, gs->ctx, gs->module,
	    &gs->sineRes, 0,
	    &gs->sineMem, sine_values, 64 * 4, "cb1");

    // init program grid
    CALprogramGrid pg = {
	.func = gs->entry,
	.gridBlock = { .width = threads_per_grp, .height = 1, .depth = 1 },
	.gridSize = { .width = gs->nr_simds, .height = 1, .depth = 1 },
	.flags = 0
    };
    gs->pg = pg;
    gs->e = 0;
    gs->have_run = false;
    gs->last_mhashpsec = 0;
    gs->nr_elms_done = 0;
    gs->all_done = false;
}

/**
 * Returns true iff the threads are currently running. Returns false
 * if they have never been started of if they completed work.
 */
bool threads_running(gpu_state_t *gs)
{
    if (!gs->have_run)
	// threads have never been started
	return false;
    CALresult res;
    res = calCtxIsEventDone(gs->ctx, gs->e);
    if (res == CAL_RESULT_OK)
      {
	gettimeofday(&gs->tv_end, NULL);
	return false;
      }
    else if (res != CAL_RESULT_PENDING)
	fatal("calCtxIsEventDone");
    return true;
}

/**
 * Analyze current results from the global buffer (if threads have
 * run at least once). And prepare next batch of work.
 *
 * Returns true iff a message hashing to the target hash has been found.
 */
bool threads_analyze_and_prepare(CALuint devi, gpu_state_t *gs)
{
    // map
    uint8_t *ptr = 0;
    CALuint pitch = 0;
    if (CAL_RESULT_OK != calResMap((CALvoid**)&ptr, &pitch, gs->globalRes, 0))
	fatal("calResMap");
    // analyze results if we have some, ie. if the threads have been started
    if (gs->have_run)
      {
	show_stats(devi, gs);
	if (verbose > 1)
	    printf(" Global buffer for first and last threads:\n");
      }
    for (int t = 0; t < gs->nr_threads; t++)
      {
	thread_state_t *ts = (thread_state_t *)ptr + t;
	if (gs->have_run && verbose > 1 && (t == 0 || t == gs->nr_threads - 1))
	    printf("  thread %d:\n", t);
	for (int e = 0; e < ELM_PER_THREAD; e++)
	  {
	    elm_state_t *elm = (elm_state_t *)ts + e;
	    if (gs->have_run && verbose > 1 &&
		    (t == 0 || t == gs->nr_threads - 1))
		printf("    elm %d: %02x(%02x%02x%02x) "
			"%02x%02x %02x%02x %02x%02x %02x%02x\n",
			e, elm->status,
			elm->_unused[0], elm->_unused[1], elm->_unused[2],
			(uint8_t)(elm->pos_wuid >> 0*8),
			(uint8_t)(elm->pos_wuid >> 1*8),
			(uint8_t)(elm->pos_wuid >> 2*8),
			(uint8_t)(elm->pos_wuid >> 3*8),
			(uint8_t)(elm->pos_wuid >> 4*8),
			(uint8_t)(elm->pos_wuid >> 5*8),
			(uint8_t)(elm->pos_wuid >> 6*8),
			(uint8_t)(elm->pos_wuid >> 7*8));
	    bool ready_for_new_wu = false;
	    if (!gs->have_run)
		ready_for_new_wu = true;
	    else
	      {
		if (elm->status == s_searching)
		    // still searching this work unit
		    ready_for_new_wu = false;
		else if (elm->status == s_found)
		  {
		    // Found a potential candidate message. Note that when
		    // found, the GPU kernel thread has already incremented
		    // the position in the work unit, so validate the position
		    // "minus 1".
		    if (verbose > 1)
			printf("Candidate found by GPU %d thread %d elm %d\n",
				devi, t, e);
		    ready_for_new_wu = false;
		    if (validate_candidate(devi, t, e, elm->pos_wuid - 1))
			// candidate was found!
			return true;
		    else
		      {
			// Candidate was a false positive. Here we need
			// special handling if it was the last candidate of the
			// work unit range
			if (!(elm->pos_wuid & ((1 << pos_lsb) - 1)))
			  {
			    printf("Special handling and end of work unit "
				    "for thread %d elm %d\n", t, e);
			    ready_for_new_wu = true;
			  }
		      }
		  }
		else if (elm->status == s_finished)
		    // finished this work unit, try next one
		    ready_for_new_wu = true;
		else if (elm->status != s_nomore)
		    fprintf(stderr, "*bug*: invalid status for thread %d "
			    "elm %d: %02x\n", t, e, elm->status), exit(1);
	      }
	    if (ready_for_new_wu)
	      {
		elm->pos_wuid = get_next_pos_wuid(&elm->status);
		// the if condition below is true only one time: the first
		// time a thread's element is detected as running out of
		// work units
		if (elm->status == s_nomore)
		    gs->nr_elms_done++;
	      }
	  }
      }
    // if after this loop we detect all threads element are done, mark
    // this device as not running anymore
    if (gs->nr_elms_done == gs->nr_threads * ELM_PER_THREAD)
	gs->all_done = true;
    // unmap
    if (CAL_RESULT_OK != calResUnmap(gs->globalRes))
	fatal("calResUnmap");
    return false;
}

void threads_start(gpu_state_t *gs)
{
    gettimeofday(&gs->tv_start, NULL);
    if (CAL_RESULT_OK != calCtxRunProgramGrid(&gs->e, gs->ctx, &gs->pg))
	fatal("calCtxRunProgram");
    if (CAL_RESULT_OK != calCtxFlush(gs->ctx))
	fatal("calCtxFlush");
    gs->have_run = true;
}

void do_run(gpu_state_t *gs_base, CALuint nr_devs)
{
    bool found = false;
    CALuint devi;
    int i = 0;
    while (!found)
      {
	if (!(i % 2000))
	    show_global_stats(gs_base, nr_devs);
	unsigned nr_all_done = 0;
	for (devi = 0; devi < nr_devs; devi++)
	  {
	    gpu_state_t *gs = gs_base + devi;
	    if (gs->all_done)
		nr_all_done++;
	    else if (!threads_running(gs))
	      {
		found = found || threads_analyze_and_prepare(devi, gs);
		if (!found && !gs->all_done)
		    threads_start(gs);
	      }
	  }
	if (nr_all_done == nr_devs)
	  {
	    printf("\nExhausted search space.\n");
	    break;
	  }
	struct timespec req = { .tv_sec = 0, .tv_nsec = 1e6 };
	nanosleep(&req, NULL);
	i++;
      }
}

void free_local_res_mem(CALcontext ctx, CALmem mem, CALresource res)
{
    if (CAL_RESULT_OK != calCtxReleaseMem(ctx, mem))
	fatal("calCtxReleaseMem");
    if (CAL_RESULT_OK != calResFree(res))
	fatal("calResFree");
}

void finish_run(gpu_state_t *gs)
{
    free_local_res_mem(gs->ctx, gs->sineMem, gs->sineRes);
    free_local_res_mem(gs->ctx, gs->msgMem, gs->msgRes);
    free_local_res_mem(gs->ctx, gs->globalMem, gs->globalRes);
    // unload module
    if (CAL_RESULT_OK != calModuleUnload(gs->ctx, gs->module))
	fatal("calModuleUnload");
    // close device
    if (CAL_RESULT_OK != calCtxDestroy(gs->ctx))
	fatal("calCtxDestroy");
    if (CAL_RESULT_OK != calDeviceClose(gs->device))
	fatal("calDeviceClose");
    // free the image and object
    if (CAL_RESULT_OK != calclFreeImage(gs->img))
	fatal("calclFreeImage");
    if (CAL_RESULT_OK != calclFreeObject(gs->obj))
	fatal("calclFreeObject");
}

void cal_puts(const CALchar *msg)
{
    fputs(msg, stdout);
}

void compile_and_run(CALuint nr_devs, const char *src)
{
    CALuint devi;
    CALdeviceattribs attribs;
    gpu_state_t *gs_base = malloc(nr_devs * sizeof (*gs_base));
    if (!gs_base)
	perror("malloc"), exit(1);
    // get attributes of the target devices, and compile for them
    printf("Launching %u threads per SIMD\n", threads_per_grp);
    for (devi = 0; devi < nr_devs; devi++)
      {
	gpu_state_t *gs = gs_base + devi;
	attribs.struct_size = sizeof(CALdeviceattribs);
	if (CAL_RESULT_OK != calDeviceGetAttribs(&attribs, devi))
	    fatal("calDeviceGetAttribs");
	printf("Device %u: %s, %u SIMDs, launching %i threads\n",
		devi, target_name(attribs.target, attribs.targetRevision),
		(unsigned)attribs.numberOfSIMD,
		attribs.numberOfSIMD * threads_per_grp);
	gs->nr_simds = attribs.numberOfSIMD;
	gs->nr_threads = gs->nr_simds * threads_per_grp;
	if (CAL_RESULT_OK != calclCompile(&gs->obj, CAL_LANGUAGE_IL, src,
		    attribs.target))
	    fatal("calclCompile");
	if (CAL_RESULT_OK != calclLink(&gs->img, &gs->obj, 1))
	    fatal("calclLink");
      }
    for (devi = 0; devi < nr_devs; devi++)
	prepare_run(devi, gs_base + devi);
    do_run(gs_base, nr_devs);
    for (devi = 0; devi < nr_devs; devi++)
	finish_run(gs_base + devi);
}

void disassemble(const char *src)
{
    CALobject obj;
    if (CAL_RESULT_OK != calclCompile(&obj, CAL_LANGUAGE_IL, src,
		disassemble_target))
	fatal("calclCompile");
    calclDisassembleObject(&obj, cal_puts);
}

void usage(const char *name)
{
    fprintf(stdout, "Usage: %s [-t <n>] [-l <n>] [-g <n>] [-v] [-d <tgt>] "
	    "[-h] [<hash>]\n"
	    "  -t <n>    Number of threads to run per SIMD (default 256)\n"
	    "  -l <n>    Attack passwords of length <n> (default 4)\n"
	    "  -g <n>    Only use first <n> GPU devices (default all)\n"
	    "  -v        Verbose mode (or -vv for very verbose)\n"
	    "  -d <tgt>  Show disassembled code for given target\n"
	    "            (<tgt> is the numerical CAL_TARGET_* value)\n"
	    "  -h        Print this help\n"
	    "  <hash>    MD5 hash to attack (default 00...00)\n"
	    , name);
}

int main(int argc, char** argv)
{
    // the CAL IL kernel makes these assumptions
    assert(sizeof (elm_state_t) == 12);
    assert(sizeof (thread_state_t) == 96);
    int opt;
    while ((opt = getopt(argc, argv, "d:g:hl:t:v")) != -1) {
	switch (opt) {
	    case 'd':
		disassemble_target = strtol(optarg, NULL, 0);
		break;
	    case 'g':
		max_gpus = strtoul(optarg, NULL, 0);
		break;
	    case 'h':
		usage(argv[0]);
		exit(0);
	    case 'l':
		pw_length = strtoul(optarg, NULL, 0);
		break;
	    case 't':
		threads_per_grp = strtoul(optarg, NULL, 0);
		break;
	    case 'v':
		verbose++;
		break;
	    default:
		usage(argv[0]);
		exit(1);
	}
    }
    if (optind < argc)
      {
	if (optind + 1 != argc)
	  {
	    fprintf(stderr, "Expected only 1 <hash> argument\n");
	    usage(argv[0]);
	    exit(1);
	  }
	load_target_values(argv[optind]);
      }
    printf("Search space: passwords of length %u\n", pw_length);
    init_md5_globals();
    if (pw_length <= 1)
      {
	printf("Note: lengths 0 and 1 are handled on the CPU.\n");
	// there is also a bug when searching these lengths on the GPU:
	// detection of exhaustion of the search space does not work well
	// when tested passwords fit entirely in the 'pos' field of pos_wuid
	for (uint64_t i = 0; i < 256; i++)
	    if (validate_candidate(0, 0, 0, i))
		exit(0);
	printf("Exhausted search space.\n");
	exit(1);
      }
    printf("Initializing CAL\n");
    if (CAL_RESULT_OK != calInit())
	fatal("calInit");
    show_ver();
    char *src;
    if (-1 == asprintf(&src, KERNEL, threads_per_grp, iterations,
		sizeof (thread_state_t) / 16,
		target_hash[0], target_hash[1], target_hash[2], target_hash[3],
		(1 << pos_lsb) - 1, s_nomore))
	perror("asprintf"), exit(1);
    if (disassemble_target != -1)
      {
	disassemble(src);
	exit(0);
      }
    CALuint nr_devs;
    if (CAL_RESULT_OK != calDeviceGetCount(&nr_devs))
	fatal("calDeviceGetCount");
    printf("Found %u device%s\n", nr_devs, nr_devs != 1 ? "s" : "");
    if (max_gpus && nr_devs > max_gpus)
      {
	nr_devs = max_gpus;
	printf("Manually limited to %u device%s\n", nr_devs,
		nr_devs != 1 ? "s" : "");
      }
    if (nr_devs >= 1)
	compile_and_run(nr_devs, src);
    calShutdown();
    free(src);
    return 0;
}
